﻿using DetourFunc;
using DetourFunc.Clr;
using System;
using System.Reflection;
using System.Runtime.InteropServices;
using WindowsPE;

namespace ConsoleApp1
{
    public delegate void SleepExDelegate(int milliseconds, bool bAlertable);

    // Install-Package DetourFunc -Version 1.0.8
    class Program
    {
        [DllImport("kernel32.dll")]
        public static extern void SleepEx(int milliseconds, bool bAlertable);

        [DllImport("kernel32", SetLastError = true, CharSet = CharSet.Ansi, BestFitMapping = false)]
        internal static extern IntPtr LoadLibrary(string lpFileName);

        [DllImport("kernel32", CharSet = CharSet.Ansi, ExactSpelling = true, SetLastError = true, BestFitMapping = false)]
        internal static extern IntPtr GetProcAddress(IntPtr hModule, string funcName);

        static void Main(string[] _)
        {
            {
                IntPtr libraryAddress = LoadLibrary("kernel32.dll");
                if (libraryAddress == IntPtr.Zero)
                {
                    return;
                }

                IntPtr funcAddr = GetProcAddress(libraryAddress, "SleepEx");
                Console.WriteLine($"GetProcAddress - Sleep API: {funcAddr.ToInt64():x}");
            }

            IntPtr sleepPtr;
            {
                sleepPtr = MethodReplacer.GetExportFunctionAddress("kernel32.dll", "SleepEx", out var _);
                //Console.WriteLine($"m64: {refAddress.ToInt64():x}");
                Console.WriteLine($"JumpTo KERNELBASE!SleepEx: {sleepPtr.ToInt64():x}");
            }

            Console.WriteLine();

            /*
            {
                IntPtr sleepEatPtr = MethodReplacer.GetExportFunctionAddressFromEAT("kernel32.dll", "SleepEx", out IntPtr refAddress);
                Console.WriteLine($"{refAddress.ToInt64():x}");
                Console.WriteLine($"{sleepEatPtr.ToInt64():x}");
            }

            Console.ReadLine();
            */

            IntPtr ptrBodyReplaceMethod;

            {
                SleepExDelegate action = Replaced_TestMethod;
                MethodDesc mdReplaceMethod = MethodDesc.ReadFromMethodInfo(action.Method);
                ptrBodyReplaceMethod = mdReplaceMethod.GetNativeFunctionPointer();
            }

            Console.WriteLine($"Address to be patched: {sleepPtr.ToInt64():x}");
            Console.WriteLine($"With this address: {ptrBodyReplaceMethod.ToInt64():x}");
            Console.WriteLine();

            using (var item = new TrampolinePatch<SleepExDelegate>())
            {
                item.JumpPatch(sleepPtr, ptrBodyReplaceMethod);

                Console.WriteLine("[After trampoline]");
                SleepEx(3000, false);
            }

            Console.WriteLine();
            Console.WriteLine("[Revert to original method]");
            SleepEx(3000, false);

            Console.WriteLine("Press any key to exit...");
            Console.ReadLine();
        }

        public static void Replaced_TestMethod(int milliseconds, bool bAlertable)
        {
            Console.WriteLine("Replaced_TestMethod called!");
        }
    }
}
